/*
   Multi-Texture Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform multitexturing.  This is when two or more
   images are combined together to appear as one.  To do multitexturing you must declare
   two sets of texture coordinates.  One for the tu, tv and one for the tu2, tv2.  This demo
   basically builds off of the texture mapping demo.  With a few lines of code we now have
   a new effect to play with.  If your hardware does not support the feature then you will
   get an error on start up.
*/


// Demo: Multi-Texture.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN


#include<d3dx9.h>

// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_TEX2)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);// Initialize Direct3D.
bool InitializeObject();                        // Initialize object we are displaying.
void RenderScene();                             // Render the 3D scene.
void ShutdownDirect3D();                        // Shutdown Direct3D.

LPDIRECT3D9 Direct3D_Object = NULL;             // Direct3D object.
LPDIRECT3DDEVICE9 D3D_Device = NULL;            // Device object.
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;   // Vertex buffer for hold our main object.
LPDIRECT3DTEXTURE9 Texture = NULL;              // Texture object to hold texture image.
LPDIRECT3DTEXTURE9 Texture2 = NULL;             // Texture object to hold texture image.

// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   FLOAT tu, tv;
   FLOAT tu2, tv2;
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;


         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;


            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                   // Message object.
   HWND hwnd;                 // Handle to the window.
   WNDCLASSEX windowClass;    // Window class object.
   bool done = false;         // Used to test if the app is running or not.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,               // The extended window style.
                         "UGPClass",         // Window class.
                         "Multi-Texture by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |// Window style.
                          WS_SYSMENU | WS_CLIPCHILDREN |// Window style.
                          WS_CLIPSIBLINGS,   // Window style.
                          100, 100,          // X, Y coords.
                          640, 480,          // Window size.
                          NULL,              // Handle to parent window.
                          NULL,              // Menu.
                          hInstance,         // Handle to app instance.
                          NULL);             // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // Set up Direct3D.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!", "Error!", MB_OK);
         return false;
      }

   // Clear it.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);       // Lighting off.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 0.8f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);

   // Initialize all models and geometry used in this demo.
   if(!InitializeObject()) return false;

   return true;
}


bool InitializeObject()
{
   // Square data for its position and texture coords.
	Vertex Polygon_Data[4] =
	{
		{-0.3f, 0.5f, 1.5f,  0.0, 0.0, 0.0, 0.0},
      {0.3f, 0.5f, 1.5f,   1.0, 0.0, 1.0, 0.0},
	   {-0.3f, -0.5f, 1.5f, 0.0, 1.0, 0.0, 1.0},
      {0.3f, -0.5f, 1.5f,  1.0, 1.0, 1.0, 1.0}
	};

   // Create the vertex buffer that will hold the square.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Polygon_Data), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      return false;
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Polygon_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the square's data into the vertex buffer.
   memcpy(Vertices, Polygon_Data, sizeof(Polygon_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();

   // Here we load the texture into the texture object.
   if(D3DXCreateTextureFromFile(D3D_Device, "Image1.bmp", &Texture) != D3D_OK)
      return false;

   // Here we load the texture into the texture object.
   if(D3DXCreateTextureFromFile(D3D_Device, "Image2.bmp", &Texture2) != D3D_OK)
      return false;

   return true;
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Set the texture stages for the first texture unit (image).
	   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	   D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

      D3D_Device->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
	   D3D_Device->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
	   D3D_Device->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	   D3D_Device->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

	   // Set the texture stages for the second texture unit (image).
      D3D_Device->SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	   D3D_Device->SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

      D3D_Device->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
      D3D_Device->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_MODULATE);
      D3D_Device->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	   D3D_Device->SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);

      // Here we set the textures we will be multi-texturing.
      D3D_Device->SetTexture(0, Texture);
      D3D_Device->SetTexture(1, Texture2);

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));

      // Set the vertex stream declaration.
      D3D_Device->SetFVF(D3DFVF_D3DVertex);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }

   if(Texture2 != NULL) 
      {
         Texture2->Release();
         Texture2 = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
         // Release the buffer.
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// When performing multitexturing you need two texture coordinates, one for the
// first image, and one for the second.  In this demo we only blended two texture
// together.  Try playing around with the code in the render function
// to see different results and effects.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com